/**
 * Copyright (C) 2012-2016  Dušan Vejnovič  <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.dussan.vaadin.dtable.properties;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.commons.lang3.ArrayUtils;
import org.dussan.vaadin.dtable.DTable;
import org.dussan.vaadin.dtable.config.formatting.Formatting;
import org.dussan.vaadin.dtable.processors.FormattingProcessor;
import org.dussan.vaadin.dtable.processors.SearchProcessor;
import org.dussan.vaadin.dtable.properties.filter.Filter;
import org.dussan.vaadin.dtable.properties.filter.FilterHelper;
import org.dussan.vaadin.dtable.properties.order.Order;
import org.dussan.vaadin.dtable.properties.order.OrderHelper;
import org.dussan.vaadin.dtable.properties.order.OrderType;
import org.dussan.vaadin.dtable.properties.search.QuickSearchHelper;
import org.dussan.vaadin.dtable.properties.search.Search;
import org.dussan.vaadin.dtable.properties.search.SearchHelper;

public class TableProperties
    implements Serializable, OrderHelper, FilterHelper, QuickSearchHelper, SearchHelper
{

    private enum ColumnStatus
    {
        HIDDEN, FROZEN, FROZEN_AND_HIDDEN, REMOVED, UNDEFINED, UNVISIBLE, VISIBLE;
    };

    private static final long serialVersionUID = 609108827865239112L;

    private long allRows = 0;

    private int left = 0;

    private int top = 0;

    private int firstColumn = 0;

    private int firstRow = 0;

    private int tableWidth = 0;

    private int screenWidth = 0;

    private String translatedWordIsEmpty = null;

    private String translatedWordIsNotEmpty = null;

    private transient List<List<Object>> originTableData = new ArrayList<>();

    private List<String> columnsHeader = new ArrayList<>();

    private Map<Integer, String> headersTooltip = new HashMap<>();

    private Map<Integer, ColumnStatus> columnsStatus = new HashMap<>();

    private transient Map<Integer, List<Object>> columnsObjects = new HashMap<>();

    private String quickSearch = new String();

    private String lastQuickSearch = new String();

    private Search lastSearch = null;

    private Filter lastFilter = null;

    private String getTranslatedWordIsEmpty()
    {
        return translatedWordIsEmpty;
    }

    public void setTranslatedWordIsEmpty( String translatedWordIsEmpty )
    {
        this.translatedWordIsEmpty = translatedWordIsEmpty;
    }

    private String getTranslatedWordIsNotEmpty()
    {
        return translatedWordIsNotEmpty;
    }

    public void setTranslatedWordIsNotEmpty( String translatedWordIsNotEmpty )
    {
        this.translatedWordIsNotEmpty = translatedWordIsNotEmpty;
    }

    public Object[] getColumnsDataType()
    {
        List<Object> dataTypes = new ArrayList<>();
        for ( List<Object> columnData : originTableData )
        {
            Object type = null;
            for ( Object data : columnData )
            {
                if ( type == null )
                {
                    type = data.getClass();
                }
                else
                {
                    break;
                }
            }
            dataTypes.add( type );
        }
        return dataTypes.toArray();
    }

    public Object getOriginTableData( int column, int row )
    {
        return originTableData.get( column )
                              .get( row );
    }

    public List<Object> getOriginTableData( int column )
    {
        return originTableData.get( column );
    }

    public void resetOriginTableData()
    {
        originTableData.clear();
    }

    public boolean isOriginTableDataEmpty()
    {
        return originTableData.isEmpty();
    }

    public void setOriginTableData( List<Object> originTableData )
    {
        this.originTableData.add( originTableData );
    }

    public int getColumnHeaderIndex( String header )
    {
        try
        {
            return columnsHeader.indexOf( header );
        }
        catch ( Exception e )
        {
            Logger.getLogger( DTable.class.getName() )
                  .log( Level.INFO, e.getLocalizedMessage(), e );
            return -1;
        }
    }

    public String getColumnHeader( int col )
    {
        try
        {
            return columnsHeader.get( col );
        }
        catch ( Exception e )
        {
            Logger.getLogger( DTable.class.getName() )
                  .log( Level.INFO, e.getLocalizedMessage(), e );
            return null;
        }
    }

    public void setColumnHeader( int column, String header )
    {
        if ( header != null )
        {
            columnsHeader.add( column, header );
            if ( !columnsStatus.containsKey( column ) )
            {
                setColumnStatus( column, ColumnStatus.UNVISIBLE );
            }
        }
    }

    public String[] getCompleteColumnsHeader()
    {
        try
        {
            return columnsHeader.toArray( new String[] {} );
        }
        catch ( Exception e )
        {
            Logger.getLogger( DTable.class.getName() )
                  .log( Level.INFO, e.getLocalizedMessage(), e );
            return null;
        }
    }

    public String[] getColumnsHeader( int[] columns )
    {
        try
        {
            List<String> headers = new ArrayList<>();
            for ( int column : columns )
            {
                if ( !isRemovedColumn( column ) && !isHiddenColumn( column ) )
                {
                    headers.add( columnsHeader.get( column ) );
                }
            }
            return headers.toArray( new String[] {} );
        }
        catch ( Exception e )
        {
            Logger.getLogger( DTable.class.getName() )
                  .log( Level.INFO, e.getLocalizedMessage(), e );
            return null;
        }
    }

    public String[] getColumnsHeader()
    {
        try
        {
            List<String> headers = new ArrayList<>();
            for ( int column = 0; column < columnsHeader.size(); column++ )
            {
                if ( !isRemovedColumn( column ) && !isHiddenColumn( column ) )
                {
                    headers.add( columnsHeader.get( column ) );
                }
            }
            return headers.toArray( new String[] {} );
        }
        catch ( Exception e )
        {
            Logger.getLogger( DTable.class.getName() )
                  .log( Level.INFO, e.getLocalizedMessage(), e );
            return null;
        }
    }

    public void setColumnsHeader( String... headers )
    {
        if ( headers != null && headers.length > 0 )
        {
            int column = columnsStatus.size();
            for ( String header : headers )
            {
                columnsHeader.add( header );
                setColumnStatus( column++, ColumnStatus.UNVISIBLE );
            }
        }
    }

    public String getHeaderTooltip( int col )
    {
        try
        {
            if ( headersTooltip.containsKey( col ) )
            {
                return headersTooltip.get( col );
            }
            else
            {
                return "";
            }
        }
        catch ( Exception e )
        {
            Logger.getLogger( DTable.class.getName() )
                  .log( Level.INFO, e.getLocalizedMessage(), e );
            return "";
        }
    }

    public void setHeaderTooltip( int column, String tooltip )
    {
        if ( tooltip != null )
        {
            headersTooltip.put( column, tooltip );
        }
    }

    public String[] getHeadersTooltip( int[] columns )
    {
        try
        {
            List<String> tooltips = new ArrayList<>();
            for ( int column : columns )
            {
                if ( !isRemovedColumn( column ) && !isHiddenColumn( column ) )
                {
                    tooltips.add( getHeaderTooltip( column ) );
                }
            }
            return tooltips.toArray( new String[] {} );
        }
        catch ( Exception e )
        {
            Logger.getLogger( DTable.class.getName() )
                  .log( Level.INFO, e.getLocalizedMessage(), e );
            return null;
        }
    }

    private boolean hasAnyColumnStatus( ColumnStatus status )
    {
        return columnsStatus.containsValue( status );
    }

    private ColumnStatus getColumnStatus( int column )
    {
        if ( columnsStatus.containsKey( column ) )
        {
            return columnsStatus.get( column );
        }
        return ColumnStatus.UNDEFINED;
    }

    private void setColumnStatus( int column, ColumnStatus columnStatus )
    {
        columnsStatus.put( column, columnStatus );
    }

    private boolean hasAnyColumnsObjects()
    {
        return !columnsObjects.isEmpty();
    }

    private void removeColumnObjects( int column )
    {
        if ( hasAnyColumnsObjects() )
        {
            if ( hasColumnObjects( column ) )
            {
                columnsObjects.remove( column );
            }
        }
    }

    private List<Integer> getColumnsObjectsKeys()
    {
        if ( hasAnyColumnsObjects() )
        {
            List<Integer> columns = new ArrayList<>( columnsObjects.keySet() );
            Collections.sort( columns );
            return columns;
        }
        return new ArrayList<>();
    }

    private boolean hasColumnObjects( int column )
    {
        if ( hasAnyColumnsObjects() )
        {
            return columnsObjects.containsKey( column );
        }
        return false;
    }

    private List<Object> getColumnObjects( int column )
    {
        if ( hasAnyColumnsObjects() )
        {
            if ( hasColumnObjects( column ) )
            {
                return columnsObjects.get( column );
            }
        }
        return new ArrayList<>();
    }

    private void setColumnObjects( int column, List<Object> columnObject )
    {
        if ( columnObject != null && !columnObject.isEmpty() )
        {
            columnsObjects.put( column, columnObject );
        }
        else if ( hasAnyColumnsObjects() )
        {
            columnsObjects.remove( column );
        }
    }

    private void resetVisibleColumns()
    {
        for ( int column = 0; column < getAllColumns(); column++ )
        {
            switch ( getColumnStatus( column ) )
            {
                case VISIBLE:
                    setColumnStatus( column, ColumnStatus.UNVISIBLE );
                    break;
                default:
                    break;
            }
        }
    }

    private int getVisibleColumnsSize()
    {
        int size = 0;
        for ( int column = 0; column < getAllColumns(); column++ )
        {
            switch ( getColumnStatus( column ) )
            {
                case VISIBLE:
                    size += 1;
                    break;
                default:
                    break;
            }
        }
        return size;
    }

    public void checkVisibleColumns( int tableWidth, int left, int top, int bodyRowHeight, int[] columnWidths,
                                     boolean enableAllColumns )
    {
        this.screenWidth = tableWidth;
        this.left = left;
        this.top = top;
        calculateLeftAndTopValues( bodyRowHeight, columnWidths );
        checkVisibleColumns( tableWidth, columnWidths, enableAllColumns );
    }

    private void calculateLeftAndTopValues( int bodyRowHeight, int[] columnWidths )
    {
        int width = 0;
        for ( int column = 0; column < getAllColumns(); column++ )
        {
            if ( !isRemovedColumn( column ) && !isHiddenColumn( column ) && !isFrozenColumn( column ) )
            {
                if ( ( width + columnWidths[column] ) > left )
                {
                    firstColumn = column;
                    left = width;
                    break;
                }
                width += columnWidths[column];
            }
        }

        firstRow = top / bodyRowHeight;
        top = firstRow * bodyRowHeight;
    }

    public void checkVisibleColumns( int tableWidth, int[] columnWidths, boolean enableAllColumns )
    {
        if ( columnWidths != null && 0 < columnWidths.length )
        {
            resetVisibleColumns();
            firstColumn = enableAllColumns ? 0 : firstColumn;

            int width = 0;
            int tempTableWidth = tableWidth + columnWidths[firstColumn];
            for ( int column = firstColumn; column < getAllColumns(); column++ )
            {
                if ( !isRemovedColumn( column ) && !isHiddenColumn( column ) && !isFrozenColumn( column ) )
                {
                    if ( !enableAllColumns )
                    {
                        setColumnStatus( column, ColumnStatus.VISIBLE );
                        if ( tempTableWidth < ( width += columnWidths[column] ) )
                        {
                            break;
                        }
                    }
                    else
                    {
                        setColumnStatus( column, ColumnStatus.VISIBLE );
                    }
                }
            }

            if ( !enableAllColumns && width < tempTableWidth )
            {
                tempTableWidth -= columnWidths[firstColumn];
                for ( int column = firstColumn - 1; 0 <= column; column-- )
                {
                    if ( !isRemovedColumn( column ) && !isHiddenColumn( column ) && !isFrozenColumn( column ) )
                    {
                        setColumnStatus( column, ColumnStatus.VISIBLE );
                        width += columnWidths[column];
                        left -= columnWidths[column];
                        if ( tempTableWidth < width )
                        {
                            firstColumn = column;
                            break;
                        }
                    }
                }
            }
        }
    }

    public void calculateTableWidth( int[] widths )
    {
        tableWidth = 0;
        for ( int column = 0; column < getAllColumns(); column++ )
        {
            if ( !isRemovedColumn( column ) && !isHiddenColumn( column ) && !isFrozenColumn( column ) )
            {
                tableWidth += widths[column];
            }
        }
    }

    public int getTableWidth()
    {
        return tableWidth;
    }

    public int getScreenWidth()
    {
        return screenWidth;
    }

    public int getLeft()
    {
        return left;
    }

    public int getTop()
    {
        return top;
    }

    public int getFirstRow()
    {
        return firstRow;
    }

    public void setFirstRow( int firstRow )
    {
        this.firstRow = firstRow;
    }

    public boolean isAllColumnsVisible()
    {
        return getAllColumns() == ( getRemovedColumnsSize() + getFrozenColumnsSize() + getHiddenColumnsSize()
            + getVisibleColumnsSize() );
    }

    public int getAllColumns()
    {
        return columnsHeader.size();
    }

    public long getAllRows()
    {
        return allRows;
    }

    public void setAllRows( long allRows )
    {
        this.allRows = allRows >= 0 ? allRows : 0;
    }

    public int[] getFrozenAndVisibleColumns()
    {
        List<Integer> columns = new ArrayList<>();
        for ( int column = 0; column < getAllColumns(); column++ )
        {
            switch ( getColumnStatus( column ) )
            {
                case FROZEN:
                case VISIBLE:
                    columns.add( column );
                    break;
                default:
                    break;
            }
        }
        return ArrayUtils.toPrimitive( columns.toArray( new Integer[] {} ) );
    }

    public int[] getFrozenAndAllVisibleColumns()
    {
        List<Integer> columns = new ArrayList<>();
        for ( int column = 0; column < getAllColumns(); column++ )
        {
            switch ( getColumnStatus( column ) )
            {
                case FROZEN:
                case UNVISIBLE:
                case VISIBLE:
                    columns.add( column );
                    break;
                default:
                    break;
            }
        }
        return ArrayUtils.toPrimitive( columns.toArray( new Integer[] {} ) );
    }

    public int[] getVisibleColumns()
    {
        List<Integer> columns = new ArrayList<>();
        for ( int column = 0; column < getAllColumns(); column++ )
        {
            switch ( getColumnStatus( column ) )
            {
                case VISIBLE:
                    columns.add( column );
                    break;
                default:
                    break;
            }
        }
        return ArrayUtils.toPrimitive( columns.toArray( new Integer[] {} ) );
    }

    public void resetRemovedColumns()
    {
        for ( int column = 0; column < getAllColumns(); column++ )
        {
            switch ( getColumnStatus( column ) )
            {
                case REMOVED:
                    setColumnStatus( column, ColumnStatus.UNVISIBLE );
                    break;
                default:
                    break;
            }
        }
    }

    public boolean isNoRemovedColumn()
    {
        return !hasAnyColumnStatus( ColumnStatus.REMOVED );
    }

    public boolean isRemovedColumn( int column )
    {
        switch ( getColumnStatus( column ) )
        {
            case REMOVED:
                return true;
            default:
                return false;
        }
    }

    private int getRemovedColumnsSize()
    {
        int size = 0;
        for ( int column = 0; column < getAllColumns(); column++ )
        {
            switch ( getColumnStatus( column ) )
            {
                case REMOVED:
                    size += 1;
                    break;
                default:
                    break;
            }
        }
        return size;
    }

    public int[] getRemovedColumns()
    {
        List<Integer> columns = new ArrayList<>();
        for ( int column = 0; column < getAllColumns(); column++ )
        {
            switch ( getColumnStatus( column ) )
            {
                case REMOVED:
                    columns.add( column );
                    break;
                default:
                    break;
            }
        }
        return ArrayUtils.toPrimitive( columns.toArray( new Integer[] {} ) );
    }

    public void setRemovedColumn( int column )
    {
        setColumnStatus( column, ColumnStatus.REMOVED );
        removeColumnObjects( column );
    }

    public void setRemovedColumns( int... columns )
    {
        for ( int column : columns )
        {
            setRemovedColumn( column );
        }
    }

    public void resetHiddenColumns()
    {
        for ( int column = 0; column < getAllColumns(); column++ )
        {
            switch ( getColumnStatus( column ) )
            {
                case FROZEN_AND_HIDDEN:
                    setColumnStatus( column, ColumnStatus.FROZEN );
                    break;
                case HIDDEN:
                    setColumnStatus( column, ColumnStatus.UNVISIBLE );
                    break;
                default:
                    break;
            }
        }
    }

    public boolean isNoHiddenColumn()
    {
        return !( hasAnyColumnStatus( ColumnStatus.FROZEN_AND_HIDDEN ) || hasAnyColumnStatus( ColumnStatus.HIDDEN ) );
    }

    public boolean isHiddenColumn( int column )
    {
        switch ( getColumnStatus( column ) )
        {
            case FROZEN_AND_HIDDEN:
            case HIDDEN:
                return true;
            default:
                return false;
        }
    }

    private int getHiddenColumnsSize()
    {
        int size = 0;
        for ( int column = 0; column < getAllColumns(); column++ )
        {
            switch ( getColumnStatus( column ) )
            {
                case FROZEN_AND_HIDDEN:
                case HIDDEN:
                    size += 1;
                    break;
                default:
                    break;
            }
        }
        return size;
    }

    public void removeHiddenColumn( int column )
    {
        switch ( getColumnStatus( column ) )
        {
            case FROZEN_AND_HIDDEN:
                setColumnStatus( column, ColumnStatus.FROZEN );
                break;
            case HIDDEN:
                setColumnStatus( column, ColumnStatus.UNVISIBLE );
                break;
            default:
                break;
        }
    }

    public boolean isHiddenColumnFrozen( int column )
    {
        switch ( getColumnStatus( column ) )
        {
            case FROZEN_AND_HIDDEN:
                return true;
            default:
                return false;
        }
    }

    public void setHiddenColumn( int column )
    {
        switch ( getColumnStatus( column ) )
        {
            case FROZEN:
                setColumnStatus( column, ColumnStatus.FROZEN_AND_HIDDEN );
                break;
            default:
                setColumnStatus( column, ColumnStatus.HIDDEN );
                break;
        }
        removeColumnObjects( column );
    }

    public int[] getHiddenColumns()
    {
        List<Integer> columns = new ArrayList<>();
        for ( int column = 0; column < getAllColumns(); column++ )
        {
            switch ( getColumnStatus( column ) )
            {
                case FROZEN_AND_HIDDEN:
                case HIDDEN:
                    columns.add( column );
                    break;
                default:
                    break;
            }
        }
        return ArrayUtils.toPrimitive( columns.toArray( new Integer[] {} ) );
    }

    public int[] getHiddenFrozenColumns()
    {
        List<Integer> columns = new ArrayList<>();
        for ( int column = 0; column < getAllColumns(); column++ )
        {
            switch ( getColumnStatus( column ) )
            {
                case FROZEN_AND_HIDDEN:
                    columns.add( column );
                    break;
                default:
                    break;
            }
        }
        return ArrayUtils.toPrimitive( columns.toArray( new Integer[] {} ) );
    }

    public String[] getHiddenColumnsString()
    {
        List<String> columns = new ArrayList<>();
        for ( int column = 0; column < getAllColumns(); column++ )
        {
            switch ( getColumnStatus( column ) )
            {
                case FROZEN_AND_HIDDEN:
                    columns.add( "-" + column );
                    break;
                case HIDDEN:
                    columns.add( "+" + column );
                    break;
                default:
                    break;
            }
        }
        return columns.toArray( new String[] {} );
    }

    public void setHiddenColumns( int... columns )
    {
        for ( int column : columns )
        {
            setHiddenColumn( column );
        }
    }

    public void resetFrozenColumns()
    {
        for ( int column = 0; column < getAllColumns(); column++ )
        {
            switch ( getColumnStatus( column ) )
            {
                case FROZEN_AND_HIDDEN:
                    setColumnStatus( column, ColumnStatus.HIDDEN );
                    break;
                case FROZEN:
                    setColumnStatus( column, ColumnStatus.UNVISIBLE );
                    break;
                default:
                    break;
            }
        }
    }

    public boolean isNoFrozenColumn()
    {
        return !hasAnyColumnStatus( ColumnStatus.FROZEN );
    }

    public boolean isFrozenColumn( int column )
    {
        switch ( getColumnStatus( column ) )
        {
            case FROZEN:
                return true;
            default:
                return false;
        }
    }

    private int getFrozenColumnsSize()
    {
        int size = 0;
        for ( int column = 0; column < getAllColumns(); column++ )
        {
            switch ( getColumnStatus( column ) )
            {
                case FROZEN:
                    size += 1;
                    break;
                default:
                    break;
            }
        }
        return size;
    }

    public boolean isAnyFrozenColumnHidden()
    {
        return hasAnyColumnStatus( ColumnStatus.FROZEN_AND_HIDDEN );
    }

    public void removeFrozenColumn( int column )
    {
        switch ( getColumnStatus( column ) )
        {
            case FROZEN_AND_HIDDEN:
                setColumnStatus( column, ColumnStatus.HIDDEN );
                break;
            case FROZEN:
                setColumnStatus( column, ColumnStatus.UNVISIBLE );
                break;
            default:
                break;
        }
    }

    public void setFrozenColumn( int column )
    {
        switch ( getColumnStatus( column ) )
        {
            case HIDDEN:
                setColumnStatus( column, ColumnStatus.FROZEN_AND_HIDDEN );
                break;
            default:
                setColumnStatus( column, ColumnStatus.FROZEN );
                break;
        }
    }

    public int[] getFrozenColumns()
    {
        List<Integer> columns = new ArrayList<>();
        for ( int column = 0; column < getAllColumns(); column++ )
        {
            switch ( getColumnStatus( column ) )
            {
                case FROZEN:
                    columns.add( column );
                    break;
                default:
                    break;
            }
        }
        return ArrayUtils.toPrimitive( columns.toArray( new Integer[] {} ) );
    }

    public void setFrozenColumns( int... columns )
    {
        for ( int column : columns )
        {
            setFrozenColumn( column );
        }
    }

    private void replaceOrderColumn( Order order )
    {
        int column = order.getColumn();
        if ( isOrdered( column ) )
        {
            removeOrder( column );
        }
        List<Object> objects = getColumnObjects( column );
        objects.add( order );
        setColumnObjects( column, objects );
    }

    @Override
    public void resetOrders()
    {
        if ( hasAnyColumnsObjects() )
        {
            for ( int column : getColumnsObjectsKeys() )
            {
                if ( isOrdered( column ) )
                {
                    removeOrder( column );
                }
            }
        }
    }

    @Override
    public boolean isNoOrders()
    {
        if ( hasAnyColumnsObjects() )
        {
            for ( int column : getColumnsObjectsKeys() )
            {
                if ( hasColumnObjects( column ) )
                {
                    for ( Object object : getColumnObjects( column ) )
                    {
                        if ( object instanceof Order )
                        {
                            return false;
                        }
                    }
                }
            }
        }
        return true;
    }

    @Override
    public void removeOrder( int column )
    {
        if ( hasAnyColumnsObjects() )
        {
            if ( isOrdered( column ) )
            {
                List<Object> objects = getColumnObjects( column );
                for ( int object = objects.size() - 1; 0 <= object; object-- )
                {
                    if ( objects.get( object ) instanceof Order )
                    {
                        objects.remove( object );
                        break;
                    }
                }
                setColumnObjects( column, objects );
            }
        }
    }

    @Override
    public boolean isOrdered( int column )
    {
        if ( hasAnyColumnsObjects() )
        {
            if ( !isNoOrders() && hasColumnObjects( column ) )
            {
                for ( Object object : getColumnObjects( column ) )
                {
                    if ( object instanceof Order )
                    {
                        return true;
                    }
                }
            }
        }
        return false;
    }

    @Override
    public Order getOrder( int column )
    {
        if ( !isNoOrders() && isOrdered( column ) )
        {
            for ( Object object : getColumnObjects( column ) )
            {
                if ( object instanceof Order )
                {
                    return (Order) object;
                }
            }
        }
        return null;
    }

    @Override
    public void setOrder( int column, OrderType type )
    {
        if ( type != null )
        {
            replaceOrderColumn( new Order( column, null, type ) );
        }
    }

    @Override
    public Order[] getOrders()
    {
        List<Order> order = new ArrayList<>();
        if ( !isNoOrders() )
        {
            for ( int column : getColumnsObjectsKeys() )
            {
                if ( isOrdered( column ) )
                {
                    order.add( getOrder( column ) );
                }
            }
        }
        return order.toArray( new Order[] {} );
    }

    private String[] getOrderColumnsString()
    {
        List<String> orderString = new ArrayList<>();
        if ( !isNoOrders() )
        {
            for ( int column : getColumnsObjectsKeys() )
            {
                if ( isOrdered( column ) )
                {
                    orderString.add( getOrder( column ).toString() );
                }
            }
            return orderString.toArray( new String[] {} );
        }
        return new String[] { "none" };
    }

    @Override
    public void setOrders( OrderType type, int... columns )
    {
        if ( type != null && columns != null && columns.length > 0 )
        {
            for ( int column : columns )
            {
                replaceOrderColumn( new Order( column, null, type ) );
            }
        }
    }

    @Override
    public void setOrders( Order[] orders )
    {
        if ( orders != null && orders.length > 0 )
        {
            for ( Order order : orders )
            {
                replaceOrderColumn( order );
            }
        }
    }

    private void newFilter( Filter filter )
    {
        int column = filter.getColumn();
        if ( isFiltered( column ) )
        {
            lastFilter = getFilter( column );
            removeFilter( column );
        }
        else
        {
            lastFilter = new Filter( column, null, null, null, null );
        }
        List<Object> objects = getColumnObjects( column );
        objects.add( filter );
        setColumnObjects( column, objects );
    }

    @Override
    public void removeFilter( int column )
    {
        if ( hasAnyColumnsObjects() )
        {
            if ( isFiltered( column ) )
            {
                List<Object> objects = getColumnObjects( column );
                for ( int object = objects.size() - 1; 0 <= object; object-- )
                {
                    if ( objects.get( object ) instanceof Filter
                        && ( (Filter) objects.get( object ) ).getColumn() >= 0 )
                    {
                        objects.remove( object );
                        break;
                    }
                }
                setColumnObjects( column, objects );
            }
        }
    }

    @Override
    public void resetFilters( String newDataType )
    {
        if ( "newFilter".equals( newDataType ) )
        {
            int column = lastFilter.getColumn();
            removeFilter( column );
            if ( lastFilter.getHeader() != null )
            {
                List<Object> objects = getColumnObjects( column );
                objects.add( lastFilter );
                setColumnObjects( column, objects );
            }
            lastFilter = null;
        }
    }

    @Override
    public boolean isNoFilters()
    {
        if ( hasAnyColumnsObjects() )
        {
            for ( int column : getColumnsObjectsKeys() )
            {
                if ( hasColumnObjects( column ) )
                {
                    for ( Object object : getColumnObjects( column ) )
                    {
                        if ( object instanceof Filter && ( (Filter) object ).getColumn() >= 0 )
                        {
                            return false;
                        }
                    }
                }
            }
        }
        return true;
    }

    @Override
    public boolean isFiltered( int column )
    {
        if ( !isNoFilters() && hasColumnObjects( column ) )
        {
            for ( Object object : getColumnObjects( column ) )
            {
                if ( object instanceof Filter && ( (Filter) object ).getColumn() >= 0 )
                {
                    return true;
                }
            }
        }
        return false;
    }

    @Override
    public Filter getFilter( int column )
    {
        if ( !isNoFilters() && isFiltered( column ) )
        {
            for ( Object object : getColumnObjects( column ) )
            {
                if ( object instanceof Filter && ( (Filter) object ).getColumn() >= 0 )
                {
                    return (Filter) object;
                }
            }
        }
        return null;
    }

    @Override
    public Filter[] getFilters()
    {
        List<Filter> filter = new ArrayList<>();
        if ( !isNoFilters() )
        {
            for ( int column : getColumnsObjectsKeys() )
            {
                if ( isFiltered( column ) )
                {
                    filter.add( getFilter( column ) );
                }
            }
        }
        return filter.toArray( new Filter[] {} );
    }

    @Override
    public void setFilter( Filter newFilter )
    {
        if ( newFilter != null )
        {
            newFilter( newFilter );
        }
    }

    @Override
    public void resetQuickSearch( String newDataObject )
    {
        if ( "quickSearch".equals( newDataObject ) )
        {
            quickSearch = lastQuickSearch;
            lastQuickSearch = new String();
        }
    }

    @Override
    public Filter[] getQuickSearchFormatting( Formatting[] formatting, boolean enableFormatting,
                                              boolean enableFilteredFrozenTable, boolean enableFilteredTable )
    {
        if ( quickSearch != null && !quickSearch.isEmpty() )
        {
            List<Filter> filters = new ArrayList<>();
            for ( int column = 0; column < getAllColumns(); column++ )
            {
                if ( ( isFrozenColumn( column ) && enableFilteredFrozenTable )
                    || ( !isFrozenColumn( column ) && enableFilteredTable ) )
                {
                    String quickSearch = this.quickSearch;
                    if ( enableFormatting )
                    {
                        for ( Formatting format : formatting )
                        {
                            if ( format.getColumn()
                                       .equals( column ) )
                            {
                                Object search = FormattingProcessor.partialProcess( format, this.quickSearch );
                                quickSearch = search == null ? this.quickSearch : search.toString();
                                break;
                            }
                        }
                    }
                    quickSearch = "%" + quickSearch + "%";
                    filters.add( new Filter( column, getColumnHeader( column ), new String[] { "LIKE" },
                                             new String[] { "%" + this.quickSearch + "%" },
                                             new String[] { quickSearch } ) );
                }
            }
            return filters.toArray( new Filter[] {} );
        }
        return null;
    }

    @Override
    public String getQuickSearch()
    {
        return quickSearch;
    }

    @Override
    public void setQuickSearch( String quickSearch )
    {
        this.lastQuickSearch = this.quickSearch;
        this.quickSearch = quickSearch != null ? quickSearch : this.quickSearch;
    }

    private void newSearch( Search search )
    {
        int column = search.getColumn();
        if ( isSearched( column ) )
        {
            lastSearch = getSearch( column );
            removeSearch( column );
        }
        else
        {
            lastSearch = new Search( column, null, null, null, null, null );
        }
        List<Object> objects = getColumnObjects( column );
        objects.add( search );
        setColumnObjects( column, objects );
    }

    @Override
    public void removeSearch()
    {
        if ( !isNoSearches() )
        {
            for ( Search search : getSearches() )
            {
                removeSearch( search.getColumn() );
            }
        }
    }

    @Override
    public void removeSearch( int column )
    {
        if ( hasAnyColumnsObjects() )
        {
            if ( isSearched( column ) )
            {
                List<Object> objects = getColumnObjects( column );
                for ( int object = objects.size() - 1; 0 <= object; object-- )
                {
                    if ( objects.get( object ) instanceof Search )
                    {
                        objects.remove( object );
                        break;
                    }
                }
                setColumnObjects( column, objects );
            }
        }
    }

    @Override
    public void resetSearch( String newDataType )
    {
        if ( "search".equals( newDataType ) )
        {
            int column = lastSearch.getColumn();
            removeSearch( column );
            if ( lastSearch.getHeader() != null )
            {
                List<Object> objects = getColumnObjects( column );
                objects.add( lastSearch );
                setColumnObjects( column, objects );
            }
            lastSearch = null;
        }
    }

    @Override
    public boolean isNoSearches()
    {
        if ( hasAnyColumnsObjects() )
        {
            for ( int column : getColumnsObjectsKeys() )
            {
                if ( hasColumnObjects( column ) )
                {
                    for ( Object object : getColumnObjects( column ) )
                    {
                        if ( object instanceof Search )
                        {
                            return false;
                        }
                    }
                }
            }
        }
        return true;
    }

    @Override
    public boolean isSearched( int column )
    {
        if ( !isNoSearches() && hasColumnObjects( column ) )
        {
            for ( Object object : getColumnObjects( column ) )
            {
                if ( object instanceof Search )
                {
                    return true;
                }
            }
        }
        return false;
    }

    @Override
    public Search getSearch( int column )
    {
        if ( !isNoSearches() && isSearched( column ) )
        {
            for ( Object object : getColumnObjects( column ) )
            {
                if ( object instanceof Search )
                {
                    return (Search) object;
                }
            }
        }
        return null;
    }

    @Override
    public Search[] getSearches()
    {
        List<Search> searches = new ArrayList<>();
        if ( !isNoSearches() )
        {
            for ( int column : getColumnsObjectsKeys() )
            {
                if ( isSearched( column ) )
                {
                    searches.add( getSearch( column ) );
                }
            }
        }
        return searches.toArray( new Search[] {} );
    }

    @Override
    public void setSearches( Search newSearch )
    {
        if ( newSearch != null )
        {
            newSearch( newSearch );
        }
    }

    @Override
    public String[] getSearchString()
    {
        if ( !isNoSearches() )
        {
            List<String> searchString = new ArrayList<>();
            for ( Search search : getSearches() )
            {
                searchString.add( "" + search.getColumn() );
                String operator = search.getOperator();
                if ( operator.contains( "LIKE" ) )
                {
                    operator = operator.startsWith( "LIKE" ) ? "LIKE" : "NOT LIKE";
                    searchString.add( operator + " '" + search.getValue() + "'" );
                    searchString.add( operator + " '" + search.getLocaleOriginalValue() + "'" );
                }
                else
                {
                    searchString.add( operator + " " + search.getValue() );
                    searchString.add( operator + " " + search.getLocaleOriginalValue() );
                }
            }
            return new ArrayList<>( searchString ).toArray( new String[] {} );
        }
        return null;
    }

    public Filter[] getCompleteFilter()
    {
        if ( !isNoFilters() || !isNoSearches() )
        {
            List<Filter> filters = new ArrayList<>();
            if ( !isNoFilters() )
            {
                filters.addAll( Arrays.asList( getFilters() ) );
            }

            if ( !isNoSearches() )
            {
                if ( filters.isEmpty() )
                {
                    for ( Search search : getSearches() )
                    {
                        filters.add( SearchProcessor.search2filter( search ) );
                    }
                }
                else
                {
                    List<Filter> tempFilters = new ArrayList<>();
                    for ( Search search : getSearches() )
                    {
                        boolean filterExist = false;
                        for ( Filter filter : getFilters() )
                        {
                            if ( search.getColumn()
                                       .equals( filter.getColumn() ) )
                            {
                                List<String> operators = new ArrayList<>( Arrays.asList( filter.getOperators() ) );
                                List<String> values = new ArrayList<>( Arrays.asList( filter.getValues() ) );
                                List<String> originalValues =
                                    new ArrayList<>( Arrays.asList( filter.getOriginalValues() ) );
                                operators.add( search.getOperator() );
                                values.add( search.getValue() );
                                originalValues.add( search.getOriginalValue() );
                                Filter newFilter =
                                    new Filter( filter.getColumn(), filter.getHeader(),
                                                operators.toArray( new String[] {} ), values.toArray( new String[] {} ),
                                                originalValues.toArray( new String[] {} ) );
                                tempFilters.add( newFilter );
                                filterExist = true;
                                break;
                            }
                        }
                        if ( !filterExist )
                        {
                            tempFilters.add( SearchProcessor.search2filter( search ) );
                        }
                    }
                    for ( Filter filter : filters )
                    {
                        boolean filterExist = false;
                        for ( Filter tempFilter : tempFilters )
                        {
                            if ( filter.getColumn()
                                       .equals( tempFilter.getColumn() ) )
                            {
                                filterExist = true;
                                break;
                            }
                        }
                        if ( !filterExist )
                        {
                            tempFilters.add( filter );
                        }
                    }
                    filters = new ArrayList<>( tempFilters );
                }
            }
            return filters.toArray( new Filter[] {} );
        }
        return null;
    }

    public String[] getNewFilter()
    {
        Filter[] filters = getCompleteFilter();
        if ( filters != null )
        {
            List<String> newFilter = new ArrayList<>();
            for ( Filter filter : filters )
            {
                for ( int element = 0; element < filter.getOperators().length; element++ )
                {
                    String value = filter.getValues()[element];
                    String operator = filter.getOperators()[element];
                    newFilter.add( "" + filter.getColumn() );
                    if ( value.isEmpty() // empty value
                        && ( operator.equals( "=" ) || operator.equals( "<>" ) ) )
                    {
                        newFilter.add( operator );
                        if ( "=".equals( operator ) )
                        {
                            newFilter.add( getTranslatedWordIsEmpty() );
                        }
                        else
                        {
                            newFilter.add( getTranslatedWordIsNotEmpty() );
                        }
                    }
                    else
                    {
                        StringBuilder value1 = new StringBuilder( "_" );
                        StringBuilder value2 = new StringBuilder( value );
                        if ( operator.endsWith( "LIKE" ) )
                        {
                            value1.insert( 0, value.startsWith( "%" ) ? "%" : "" );
                            value1.append( value.endsWith( "%" ) ? "%" : "" );
                            value1.insert( 0, "'" )
                                  .append( "'" );
                            value2.insert( 0, "'" )
                                  .append( "'" );
                        }
                        value1.insert( 0, operator + " " );
                        value2.insert( 0, operator + " " );
                        newFilter.add( value1.toString() );
                        newFilter.add( value2.toString() );
                    }
                }
            }
            return newFilter.toArray( new String[] {} );
        }
        return null;
    }

    public String getFilterHtml()
    {
        Filter[] filters = getCompleteFilter();
        if ( filters != null )
        {
            StringBuilder filterString = new StringBuilder();
            for ( Filter filter : filters )
            {
                filterString.append( "<b>[" + getColumnHeader( filter.getColumn() ) + "]:</b> " );
                StringBuilder operatorString = new StringBuilder();
                for ( int operator = 0; operator < filter.getOperators().length; operator++ )
                {
                    operatorString.append( ", " );
                    if ( "=".equals( filter.getOperators()[operator] ) && filter.getValues()[operator].isEmpty() )
                    {
                        operatorString.append( getTranslatedWordIsEmpty() );
                    }
                    else if ( "<>".equals( filter.getOperators()[operator] ) && filter.getValues()[operator].isEmpty() )
                    {
                        operatorString.append( getTranslatedWordIsNotEmpty() );
                    }
                    else
                    {
                        operatorString.append( filter.getOperators()[operator] );
                        operatorString.append( " " );
                        operatorString.append( filter.getOperators()[operator].endsWith( "LIKE" ) ? "'" : "" );
                        operatorString.append( filter.getValues()[operator] );
                        operatorString.append( filter.getOperators()[operator].endsWith( "LIKE" ) ? "'" : "" );
                    }
                }
                filterString.append( operatorString.substring( 2 ) + "; " );
            }
            return filterString.toString()
                               .trim();
        }
        return null;
    }

    public int[] getHiddenAndRemovedColumns( boolean includeHiddenColumns, boolean includeRemovedColumns )
    {
        int[] removedColumns = new int[] {};
        if ( includeHiddenColumns && !isNoHiddenColumn() )
        {
            removedColumns = ArrayUtils.addAll( removedColumns, getHiddenColumns() );
        }
        if ( includeRemovedColumns && !isNoRemovedColumn() )
        {
            removedColumns = ArrayUtils.addAll( removedColumns, getRemovedColumns() );
        }
        Arrays.sort( removedColumns );
        return removedColumns;
    }

    public Map<String, String> getData()
    {
        Map<String, String> data = new HashMap<>();

        data.put( "left", Integer.toString( getLeft() ) );
        data.put( "top", Integer.toString( getTop() ) );
        data.put( "firstRow", Integer.toString( getFirstRow() ) );
        data.put( "allColumns", Integer.toString( getAllColumns() ) );
        data.put( "allRows", Long.toString( getAllRows() ) );
        data.put( "visibleColumns", Arrays.toString( getVisibleColumns() ) );
        data.put( "allColumnsVisible", Boolean.toString( isAllColumnsVisible() ) );
        data.put( "tableWidth", Integer.toString( getTableWidth() ) );
        data.put( "isAnyFrozenColumnHidden", Boolean.toString( isAnyFrozenColumnHidden() ) );

        if ( !isNoOrders() )
        {
            data.put( "order", Arrays.toString( getOrderColumnsString() ) );
        }
        else
        {
            data.remove( "order" );
        }

        if ( !isNoFrozenColumn() )
        {
            data.put( "frozenColumns", Arrays.toString( getFrozenColumns() ) );
        }
        else
        {
            data.remove( "frozenColumns" );
        }

        return data;
    }

}
