/**
 * Copyright (C) 2012-2016  Dušan Vejnovič  <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.dussan.vaadin.dtable.query;

import java.io.Serializable;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import org.dussan.vaadin.dtable.config.formatting.Formatting;
import org.dussan.vaadin.dtable.properties.filter.Filter;
import org.dussan.vaadin.dtable.properties.order.Order;
import org.dussan.vaadin.dtable.properties.order.OrderType;
import org.dussan.vaadin.dtable.query.elements.Field;
import org.dussan.vaadin.dtable.query.elements.Having;
import org.dussan.vaadin.dtable.query.elements.HavingType;
import org.dussan.vaadin.dtable.query.elements.OrderBy;
import org.dussan.vaadin.dtable.query.elements.OrderByType;
import org.dussan.vaadin.dtable.query.elements.Where;
import org.dussan.vaadin.dtable.query.elements.WhereType;

public class TableQuery
    implements Serializable
{

    private static final long serialVersionUID = -8049470491981015573L;

    private boolean enableCastFilters = false;

    private String castFiltersAs = "DECIMAL (10,2)";

    private WhereType filterWhereType = WhereType.AND;

    private HavingType quickSearchHavingType = HavingType.OR;

    private Query query = null;

    private Query lockedQuery = null;

    private Formatting[] formattingColumns = null;

    public TableQuery()
    {
        query = new Query();
    }

    private boolean isFormattingColumn( int column )
    {
        if ( formattingColumns != null && 0 < formattingColumns.length )
        {
            for ( Formatting formatting : formattingColumns )
            {
                if ( formatting.getColumn()
                               .equals( column ) )
                {
                    return true;
                }
            }
        }
        return false;
    }

    private boolean isNumberFormattingColumn( int column )
    {
        if ( formattingColumns != null && 0 < formattingColumns.length )
        {
            for ( Formatting formatting : formattingColumns )
            {
                if ( formatting.getColumn()
                               .equals( column ) )
                {
                    if ( formatting.getFormat() instanceof NumberFormat
                        || formatting.getFormat() instanceof DecimalFormat )
                    {
                        return true;
                    }
                }
            }
        }
        return false;
    }

    public boolean isEnabledCastFilters()
    {
        return enableCastFilters;
    }

    public void enableCastFilters( boolean enableCastFilters )
    {
        this.enableCastFilters = enableCastFilters;
    }

    public void castFiltersAs( String castFiltersAs )
    {
        if ( castFiltersAs != null && !castFiltersAs.isEmpty() )
        {
            this.castFiltersAs = castFiltersAs;
        }
    }

    public void setFilterWhereType( WhereType filterWhereType )
    {
        this.filterWhereType = filterWhereType;
    }

    public WhereType getFilterWhereType()
    {
        return filterWhereType;
    }

    public void setQuickSearchHavingType( HavingType quickSearchHavingType )
    {
        this.quickSearchHavingType = quickSearchHavingType;
    }

    public HavingType getQuickSearchHavingType()
    {
        return quickSearchHavingType;
    }

    public Query getQuery()
    {
        return lockedQuery == null ? query : lockedQuery;
    }

    public void setQuery( Query query )
    {
        this.query = query;
    }

    public Query getUnlockedQuery()
    {
        return query;
    }

    public boolean isQueryLocked()
    {
        return lockedQuery != null;
    }

    public TableQuery lockQuery()
    {
        if ( lockedQuery == null )
        {
            lockedQuery = Query.clone( query );
        }
        return this;
    }

    public TableQuery unlockQuery()
    {
        lockedQuery = null;
        return this;
    }

    public Query setProperties( Order[] orders, Filter[] filters, Filter[] quickSearch, Formatting[] formattingColumns,
                                int[] hiddenOrRemovedColumns, Object defaultValueForHiddenAndRemovedColumns,
                                boolean enabledOriginalValues )
    {
        if ( !getQuery().getFields()
                        .isEmpty() )
        {
            Map<Integer, Field> fields = getQuery().getFields();
            this.formattingColumns = formattingColumns;

            if ( orders != null && 0 < orders.length )
            {
                for ( Order order : orders )
                {
                    OrderBy orderBy = new OrderBy( fields.get( order.getColumn() )
                                                         .getAlias() );
                    if ( OrderType.ASCEDING.equals( order.getType() ) )
                    {
                        getQuery().addOrderBy( orderBy.orderBy( OrderByType.ASCENDING ) );
                    }
                    else
                    {
                        getQuery().addOrderBy( orderBy.orderBy( OrderByType.DESCENDING ) );
                    }
                }
            }

            if ( filters != null && 0 < filters.length )
            {
                List<Object> whereClause = new ArrayList<>();
                for ( Filter filter : filters )
                {
                    List<Object> where = new ArrayList<>();
                    if ( filter != null )
                    {
                        for ( int operator = 0; operator < filter.getOperators().length; operator++ )
                        {
                            if ( filter.getValues()[operator].isEmpty()
                                && filter.getOperators()[operator].equals( "=" ) )
                            {
                                if ( !isFormattingColumn( filter.getColumn() ) )
                                {
                                    Object whereEmpty = Where.isEmpty( fields.get( filter.getColumn() )
                                                                             .getName() );
                                    Object whereNull = Where.isNull( fields.get( filter.getColumn() )
                                                                           .getName() );
                                    where.add( Where.OR( whereEmpty, whereNull ) );
                                }
                                else
                                {
                                    Object whereNull = Where.isNull( fields.get( filter.getColumn() )
                                                                           .getName() );
                                    where.add( whereNull );
                                }
                            }
                            else if ( filter.getValues()[operator].isEmpty()
                                && filter.getOperators()[operator].equals( "<>" ) )
                            {
                                if ( !isFormattingColumn( filter.getColumn() ) )
                                {
                                    Object whereEmpty = Where.isNotEmpty( fields.get( filter.getColumn() )
                                                                                .getName() );
                                    Object whereNull = Where.isNotNull( fields.get( filter.getColumn() )
                                                                              .getName() );
                                    where.add( Where.AND( whereEmpty, whereNull ) );
                                }
                                else
                                {
                                    Object whereNull = Where.isNotNull( fields.get( filter.getColumn() )
                                                                              .getName() );
                                    where.add( whereNull );
                                }
                            }
                            else
                            {
                                StringBuilder whereBuilder = new StringBuilder();
                                whereBuilder.append( fields.get( filter.getColumn() )
                                                           .getName() );
                                if ( enableCastFilters && isNumberFormattingColumn( filter.getColumn() )
                                    && !filter.getOperators()[operator].endsWith( "LIKE" ) )
                                {
                                    whereBuilder =
                                        new StringBuilder( "CAST(" + whereBuilder + " AS " + castFiltersAs + ")" );
                                }
                                whereBuilder.append( " " + filter.getOperators()[operator] );
                                whereBuilder.append( " '" + filter.getOriginalValues()[operator] + "'" );
                                where.add( whereBuilder.toString() );
                            }
                        }
                        // add all where clauses
                        if ( 1 < where.size() )
                        {
                            whereClause.add( Where.AND( where.toArray() ) );
                        }
                        else if ( !where.isEmpty() )
                        {
                            whereClause.add( where.get( 0 ) );
                        }
                    }
                }

                if ( !whereClause.isEmpty() )
                {
                    if ( WhereType.AND.equals( filterWhereType ) )
                    {
                        getQuery().addWhereCondition( Where.AND( whereClause.toArray() ) );
                    }
                    else
                    {
                        getQuery().addWhereCondition( Where.OR( whereClause.toArray() ) );
                    }
                }
            }

            if ( quickSearch != null && 0 < quickSearch.length )
            {
                List<Object> havingClause = new ArrayList<>();
                for ( Filter search : quickSearch )
                {
                    if ( search != null )
                    {
                        havingClause.add( fields.get( search.getColumn() )
                                                .getName()
                            + " " + search.getOperators()[0] + " '" + search.getOriginalValues()[0] + "'" );
                    }
                }

                if ( !havingClause.isEmpty() )
                {
                    if ( HavingType.AND.equals( quickSearchHavingType ) )
                    {
                        getQuery().addHavingCondition( Having.AND( havingClause.toArray() ) );
                    }
                    else
                    {
                        getQuery().addHavingCondition( Having.OR( havingClause.toArray() ) );
                    }
                }
            }

            if ( hiddenOrRemovedColumns != null && 0 < hiddenOrRemovedColumns.length )
            {
                for ( int column : hiddenOrRemovedColumns )
                {
                    if ( fields.keySet()
                               .contains( column )
                        && !enabledOriginalValues )
                    {
                        Field field = fields.get( column );
                        fields.put( column, new Field( defaultValueForHiddenAndRemovedColumns.toString(),
                                                       field.getAlias(), field.getFunction() ) );
                        getQuery().removeGroupBy( field.getAlias() );
                        getQuery().removeOrderBy( field.getAlias() );
                    }
                }
                getQuery().addFields( fields );
            }
        }
        return getQuery();
    }

}
