/**
 * Copyright (C) 2012-2016  Dušan Vejnovič  <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.dussan.vaadin.dtable.windows.exporter;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.dussan.vaadin.dtable.DTable;
import org.dussan.vaadin.dtable.translations.TranslatedWords;
import org.dussan.vaadin.dtable.windows.exporter.base.Exporter;
import org.dussan.vaadin.dtable.windows.exporter.type.ExportedType;
import org.dussan.vaadin.dtable.windows.exporter.type.ExportingColumnType;

import com.healthmarketscience.jackcess.ColumnBuilder;
import com.healthmarketscience.jackcess.Database;
import com.healthmarketscience.jackcess.DatabaseBuilder;
import com.healthmarketscience.jackcess.Table;
import com.healthmarketscience.jackcess.TableBuilder;
import com.healthmarketscience.jackcess.impl.JetFormat;

public class MdbExporter
    extends Exporter
{

    private static final long serialVersionUID = -7064150607674977016L;

    private int columnStep = 255;

    private List<String> tableNames = null;

    private Map<Integer, Integer> textLengthOfStringColumns = null;

    private Database database = null;

    public MdbExporter( DTable dTable )
    {
        super( dTable );

        tableNames = new ArrayList<>();
        textLengthOfStringColumns = new HashMap<>();

        setCaption( getDTable().getTranslation()
                               .get( getDTable().getTableConfig()
                                                .getLocale(),
                                     TranslatedWords.EXPORTING_DATA_TO_MS_ACCESS_FILE ) );
        setExportedType( ExportedType.MS_ACCESS );
    }

    private List<Object> checkStringValuesLength( int allColumns, int allRows, List<Object> data )
    {
        if ( !textLengthOfStringColumns.isEmpty() )
        {
            try
            {
                for ( int row = 0; row < allRows; row++ )
                {
                    List<Object> values = data.subList( allColumns * row, allColumns * ( row + 1 ) );
                    for ( int column : textLengthOfStringColumns.keySet() )
                    {
                        int maxLength = textLengthOfStringColumns.get( column );
                        String value = values.get( column )
                                             .toString();
                        if ( value != null && maxLength < value.length() )
                        {
                            data.set( allColumns * row + column, value.substring( 0, maxLength ) );
                        }
                    }
                }
            }
            catch ( Exception e )
            {
                Logger.getLogger( DTable.class.getName() )
                      .log( Level.INFO, e.getLocalizedMessage(), e );
            }
        }
        return data;
    }

    private void createTablesHeader( String name )
    {
        try
        {
            textLengthOfStringColumns.clear();

            int columnsDefinitionSize = getDTable().getTableExporterConfig()
                                                   .getColumnDefinition( name )
                                                   .size();
            String partialTableName = getDTable().getTranslation()
                                                 .get( getDTable().getTableConfig()
                                                                  .getLocale(),
                                                       TranslatedWords.MS_ACCESS_PARTIAL_TABLE_NAME );
            List<Integer> columns = new ArrayList<>( getDTable().getTableExporterConfig()
                                                                .getColumnDefinition( name )
                                                                .keySet() );
            Collections.sort( columns );

            int columnStep = this.columnStep;
            if ( getDTable().getTableExporterConfig()
                            .isRowIndexEnabled() )
            {
                columnStep -= 1;
            }

            tableNames.clear();
            for ( int tableCounter = 0; tableCounter <= columnsDefinitionSize / columnStep; tableCounter++ )
            {
                String tableName = getDTable().getTableExporterConfig()
                                              .getBaseName()
                                              .equals( name ) ? getDTable().getTableExporterConfig()
                                                                           .getTableName()
                                                              : name;
                tableName += columnsDefinitionSize <= columnStep ? "" : ( partialTableName + ( tableCounter + 1 ) );
                TableBuilder tableBuilder = new TableBuilder( tableName );

                if ( getDTable().getTableExporterConfig()
                                .isRowIndexEnabled() )
                {
                    ColumnBuilder tableColumn = new ColumnBuilder( "#", ExportingColumnType.LONG.getType() );
                    tableBuilder.addColumn( tableColumn.toColumn() );
                }

                for ( int column = columnStep * tableCounter; column < columnStep * ( tableCounter + 1 )
                    && column < columnsDefinitionSize; column++ )
                {
                    String header = getDTable().getTableExporterConfig()
                                               .getColumnDefinitionHeader( name, columns.get( column ) );
                    ExportingColumnType columnExportedType = getDTable().getTableExporterConfig()
                                                                        .getColumnDefinitionExportingType( name,
                                                                                                           columns.get( column ) );
                    Integer scale = getDTable().getTableExporterConfig()
                                               .getColumnDefinitionScale( name, columns.get( column ) );
                    Integer precision = getDTable().getTableExporterConfig()
                                                   .getColumnDefinitionPrecision( name, columns.get( column ) );
                    Integer length = getDTable().getTableExporterConfig()
                                                .getColumnDefinitionLength( name, columns.get( column ) );

                    ColumnBuilder tableColumn = new ColumnBuilder( header, columnExportedType.getType() );
                    switch ( columnExportedType.getType() )
                    {
                        case DOUBLE:
                        case FLOAT:
                        case NUMERIC:
                            if ( scale != null && precision != null )
                            {
                                tableColumn.setScale( scale );
                                tableColumn.setPrecision( precision );
                            }
                            break;
                        case TEXT:
                            if ( length == null )
                            {
                                length = Integer.valueOf( JetFormat.TEXT_FIELD_MAX_LENGTH );
                            }
                            length = Math.min( length, JetFormat.TEXT_FIELD_MAX_LENGTH );
                            tableColumn.setLength( length );
                            textLengthOfStringColumns.put( column, length );
                            break;
                        default:
                            break;
                    }
                    tableBuilder.addColumn( tableColumn.toColumn() );
                }
                tableBuilder.toTable( database );
                tableNames.add( tableName );
            }
        }
        catch ( Exception e )
        {
            Logger.getLogger( DTable.class.getName() )
                  .log( Level.INFO, e.getLocalizedMessage(), e );
            setExportedStatus( EXPORT_ERROR, e.getLocalizedMessage() );
        }
    }

    @Override
    protected void exportDataIntoFile()
    {
        try
        {
            if ( 0 < getDTable().getTableExporterConfig()
                                .getColumnsDefinitionName().length )
            {
                float completeStep = 0f;
                float step = 1f / getDTable().getTableExporterConfig()
                                             .getColumnsDefinitionName().length;

                setExportedStatus( EXPORT_BEGIN );
                if ( openFile( getDTable().getTableExporterConfig()
                                          .getFilename() ) )
                {
                    for ( String name : getDTable().getTableExporterConfig()
                                                   .getColumnsDefinitionName() )
                    {
                        if ( prepareExportedData( name ) && canContinue() )
                        {
                            writeToFile( name, completeStep, step );
                            completeStep += step;
                        }
                        else
                        {
                            setExportedStatus( EXPORT_ERROR );
                        }

                        // check status
                        if ( !canContinue() )
                        {
                            closeFile();
                            if ( getExportedStatus() == EXPORT_STOP )
                            {
                                setExportedStatus( EXPORT_STOPPED );
                            }
                            return;
                        }
                    }
                }
                else
                {
                    closeFile();
                    setExportedStatus( EXPORT_ERROR );
                    return;
                }

                closeFile();
                setExportedStatus( EXPORT_COMPLETE );
            }
            else
            {
                setExportedStatus( EXPORT_ERROR, getDTable().getTranslation()
                                                            .get( getDTable().getTableConfig()
                                                                             .getLocale(),
                                                                  TranslatedWords.NO_DATA_FOR_EXPORT ) );
            }
        }
        catch ( Exception e )
        {
            Logger.getLogger( DTable.class.getName() )
                  .log( Level.INFO, e.getLocalizedMessage(), e );
            setExportedStatus( EXPORT_ERROR, e.getLocalizedMessage() );
        }
    }

    @Override
    protected boolean openFile( String name )
    {
        try
        {
            database = DatabaseBuilder.create( getDTable().getTableExporterConfig()
                                                          .getMsAccessVersion()
                                                          .getVersion(),
                                               getTempFile( name ) );
            return true;
        }
        catch ( Exception e )
        {
            Logger.getLogger( DTable.class.getName() )
                  .log( Level.INFO, e.getLocalizedMessage(), e );
            setExportedStatus( EXPORT_ERROR, e.getLocalizedMessage() );
            return false;
        }
    }

    @Override
    protected void writeToFile( String name, float completeStep, float step )
    {
        try
        {
            int columnStep = this.columnStep;
            if ( getDTable().getTableExporterConfig()
                            .isRowIndexEnabled() )
            {
                columnStep -= 1;
            }

            createTablesHeader( name );
            if ( getExportedStatus() != EXPORT_ERROR )
            {
                long allRows = getAllRows( name );
                int allColumns = getAllColumns( name );
                if ( 0 < allColumns && 0 < allRows && 0 < getExportedStep( name ) )
                {
                    int exportedStep = allRows < getExportedStep( name ) ? (int) allRows : getExportedStep( name );

                    for ( int mainRow = 0; mainRow < allRows; mainRow += exportedStep )
                    {
                        int rows =
                            (int) ( ( mainRow + exportedStep ) <= allRows ? exportedStep : ( allRows - mainRow ) );
                        List<Object[]> data = new ArrayList<>();
                        List<Object> rowsData =
                            checkStringValuesLength( allColumns, exportedStep,
                                                     Arrays.asList( getFromattedData( name, mainRow, rows ) ) );

                        for ( int tableName = 0; tableName < tableNames.size(); tableName++ )
                        {
                            int fromColumn = columnStep * tableName;
                            int toColumn = columnStep * ( tableName + 1 );
                            toColumn = Math.min( toColumn, allColumns );
                            Table table = database.getTable( tableNames.get( tableName ) );

                            data.clear();
                            for ( int row = 0; row < rows; row++ )
                            {
                                List<Object> rowData =
                                    new ArrayList<>( rowsData.subList( allColumns * row + fromColumn,
                                                                       allColumns * row + toColumn ) );
                                if ( getDTable().getTableExporterConfig()
                                                .isRowIndexEnabled() )
                                {
                                    rowData.add( 0, mainRow + row + 1 );
                                }
                                data.add( rowData.toArray() );

                                // exports was stopped
                                if ( !canContinue() )
                                {
                                    setExportedStatus( EXPORT_STOPPED );
                                    return;
                                }
                            }
                            table.addRows( data );
                        }

                        setExportedStatus( step * mainRow / allRows + completeStep );
                    }
                }
                else
                {
                    setExportedStatus( EXPORT_ERROR, getDTable().getTranslation()
                                                                .get( getDTable().getTableConfig()
                                                                                 .getLocale(),
                                                                      TranslatedWords.NO_DATA_FOR_EXPORT ) );
                }
            }
        }
        catch ( Exception e )
        {
            Logger.getLogger( DTable.class.getName() )
                  .log( Level.INFO, e.getLocalizedMessage(), e );
            setExportedStatus( EXPORT_ERROR, e.getLocalizedMessage() );
        }
    }

    @Override
    protected void closeFile()
    {
        try
        {
            database.close();
        }
        catch ( Exception e )
        {
            Logger.getLogger( DTable.class.getName() )
                  .log( Level.INFO, e.getLocalizedMessage(), e );
            setExportedStatus( EXPORT_ERROR, e.getLocalizedMessage() );
        }
    }

}
