/**
 * Copyright (C) 2012-2016  Dušan Vejnovič  <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.dussan.vaadin.dtable.windows.exporter.base;

import java.io.File;
import java.text.NumberFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.dussan.vaadin.dtable.DTable;
import org.dussan.vaadin.dtable.config.formatting.Formatting;
import org.dussan.vaadin.dtable.processors.CompressingProcessor;
import org.dussan.vaadin.dtable.processors.FormattingProcessor;
import org.dussan.vaadin.dtable.translations.TranslatedWords;
import org.dussan.vaadin.dtable.utils.ErrorUtil;
import org.dussan.vaadin.dtable.windows.exporter.type.CompressionType;
import org.dussan.vaadin.dtable.windows.exporter.type.ExportedType;
import org.dussan.vaadin.dtable.windows.exporter.type.ExportingColumnType;

import com.healthmarketscience.jackcess.impl.JetFormat;
import com.vaadin.data.Property.ValueChangeEvent;
import com.vaadin.data.Property.ValueChangeListener;
import com.vaadin.event.UIEvents.PollEvent;
import com.vaadin.event.UIEvents.PollListener;
import com.vaadin.server.FileResource;
import com.vaadin.server.Page;
import com.vaadin.ui.Alignment;
import com.vaadin.ui.Button;
import com.vaadin.ui.Button.ClickEvent;
import com.vaadin.ui.Button.ClickListener;
import com.vaadin.ui.CheckBox;
import com.vaadin.ui.HorizontalLayout;
import com.vaadin.ui.Label;
import com.vaadin.ui.UI;
import com.vaadin.ui.VerticalLayout;
import com.vaadin.ui.Window;

public abstract class Exporter
    extends Window
{

    private static final long serialVersionUID = -6034272525155648866L;

    public static final int DISABLED = -1;

    public static final int DISABLING = 200;

    public static final int EXPORT_STEP = 100;

    public static final Float EXPORT_BEGIN = 0.0f;

    public static final Float EXPORT_COMPLETE = 1.0f;

    public static final Float EXPORT_COMPLETED = 2.0f;

    public static final Float EXPORT_ERROR = -100.0f;

    public static final Float EXPORT_NOT_STARTED = -95.0f;

    public static final Float EXPORT_PREPARE = 3.0f;

    public static final Float EXPORT_PREPARED = 4.0f;

    public static final Float EXPORT_STOP = -98.0f;

    public static final Float EXPORT_STOPPED = -99.0f;

    private boolean exportInterrupted = false;

    private Float exportedStatus = EXPORT_NOT_STARTED;

    private int exportStatus = DISABLED;

    private int refreshDelay = 1000;

    private String tempFolder = null;

    private Button closeButton = null;

    private Button exportButton = null;

    private CheckBox includeHiddenColumns = null;

    private CheckBox removeOrdersFromColumns = null;

    private Label exportStatusLabel = null;

    private ExportedType exportingType = null;

    private File archivedFile = null;

    private NumberFormat exportStatusFormatter = null;

    private DTable dTable = null;

    /**
     * Create a new instance.
     */
    private Exporter()
    {
        // Auto-generated constructor stub
    }

    /**
     * Creates a new instance with dTable object.
     * 
     * @param dTable dTable instance
     */
    public Exporter( DTable dTable )
    {
        this();
        this.dTable = dTable;

        // export status formatter
        exportStatusFormatter = NumberFormat.getPercentInstance( getDTable().getTableConfig()
                                                                            .getLocale() );
        exportStatusFormatter.setMinimumFractionDigits( 2 );
        exportStatusFormatter.setMaximumFractionDigits( 2 );
        refreshDelay = Math.max( refreshDelay, getDTable().getTableExporterConfig()
                                                          .getWindowRefreshDelay() );

        // initialization
        init();
    }

    /**
     * Initialization
     */
    private void init()
    {
        center();
        setClosable( false );
        setDraggable( false );
        setModal( true );
        setResizable( false );
        setWidth( 70, Unit.PERCENTAGE );

        // check box: include hidden columns
        includeHiddenColumns = new CheckBox( dTable.getTranslation()
                                                   .get( dTable.getTableConfig()
                                                               .getLocale(),
                                                         TranslatedWords.INCLUDE_HIDDEN_COLUMNS ) );
        includeHiddenColumns.setImmediate( true );
        includeHiddenColumns.setValue( dTable.getTableExporterConfig()
                                             .isPrepareDataWithHiddenColumnsEnabled() );
        includeHiddenColumns.addValueChangeListener( new ValueChangeListener()
        {
            private static final long serialVersionUID = -2193783918369047227L;

            @Override
            public void valueChange( ValueChangeEvent event )
            {
                dTable.getTableExporterConfig()
                      .setPrepareDataWithHiddenColumnsEnabled( Boolean.getBoolean( event.getProperty()
                                                                                        .getValue()
                                                                                        .toString() ) );
            }
        } );

        // check box: remove orders
        removeOrdersFromColumns = new CheckBox( dTable.getTranslation()
                                                      .get( dTable.getTableConfig()
                                                                  .getLocale(),
                                                            TranslatedWords.REMOVE_ORDERS_FROM_COLUMNS ) );
        removeOrdersFromColumns.setImmediate( true );
        removeOrdersFromColumns.setValue( !dTable.getTableExporterConfig()
                                                 .isPrepareDataWithOrdersEnabled() );
        removeOrdersFromColumns.addValueChangeListener( new ValueChangeListener()
        {
            private static final long serialVersionUID = 5770319267885428215L;

            @Override
            public void valueChange( ValueChangeEvent event )
            {
                dTable.getTableExporterConfig()
                      .setPrepareDataWithHiddenColumnsEnabled( Boolean.getBoolean( event.getProperty()
                                                                                        .getValue()
                                                                                        .toString() ) );
            }
        } );

        // export button
        exportButton = new Button( dTable.getTranslation()
                                         .get( dTable.getTableConfig()
                                                     .getLocale(),
                                               TranslatedWords.EXPORT ) );
        exportButton.addClickListener( new ClickListener()
        {
            private static final long serialVersionUID = -2991532173977807930L;

            @SuppressWarnings( "deprecation" )
            @Override
            public void buttonClick( ClickEvent event )
            {
                // export
                if ( getDTable().getTranslation()
                                .get( dTable.getTableConfig()
                                            .getLocale(),
                                      TranslatedWords.EXPORT )
                                .equals( exportButton.getCaption() ) )
                {
                    closeButton.setEnabled( false );

                    includeHiddenColumns.setEnabled( false );
                    removeOrdersFromColumns.setEnabled( false );
                    exportStatusLabel.setValue( dTable.getTranslation()
                                                      .get( dTable.getTableConfig()
                                                                  .getLocale(),
                                                            TranslatedWords.PREPARING_DATA )
                        + "..." );
                    exportButton.setCaption( dTable.getTranslation()
                                                   .get( dTable.getTableConfig()
                                                               .getLocale(),
                                                         TranslatedWords.STOP_PREPARING_DATA ) );

                    tempFolder = System.getProperty( "java.io.tmpdir" );
                    tempFolder += Character.toString( File.separatorChar ) + UUID.randomUUID();

                    exportStatus = refreshDelay;
                    new ExportProcessRunner().start();
                    UI.getCurrent()
                      .setPollInterval( refreshDelay );
                    UI.getCurrent()
                      .addPollListener( new PollListener()
                      {
                          private static final long serialVersionUID = -4112423287151625267L;

                          @Override
                          public void poll( PollEvent event )
                          {
                              exportStatusWatcher();
                          }
                      } );
                }
                // stop preparing data
                else if ( getDTable().getTranslation()
                                     .get( dTable.getTableConfig()
                                                 .getLocale(),
                                           TranslatedWords.STOP_PREPARING_DATA )
                                     .equals( exportButton.getCaption() ) )
                {
                    exportInterrupted = true;
                    UI.getCurrent()
                      .setPollInterval( DISABLING );

                }
                // download the data
                else if ( getDTable().getTranslation()
                                     .get( dTable.getTableConfig()
                                                 .getLocale(),
                                           TranslatedWords.DOWNLOAD_THE_DATA )
                                     .equals( exportButton.getCaption() ) )
                {
                    FileResource fileResource = new FileResource( archivedFile );
                    Page.getCurrent()
                        .open( fileResource, "_top", true );
                    closeWindow();
                }
            }
        } );

        // close button
        closeButton = new Button( dTable.getTranslation()
                                        .get( dTable.getTableConfig()
                                                    .getLocale(),
                                              TranslatedWords.CLOSE ) );
        closeButton.addClickListener( new ClickListener()
        {
            private static final long serialVersionUID = 5181828293982712963L;

            @Override
            public void buttonClick( ClickEvent event )
            {
                if ( tempFolder != null )
                {
                    deleteTempFolder( new File( tempFolder ) );
                }
                closeWindow();
            }
        } );

        exportStatusLabel = new Label( dTable.getTranslation()
                                             .get( dTable.getTableConfig()
                                                         .getLocale(),
                                                   TranslatedWords.TO_EXPORT_THE_DATA_PRESS_THE_BUTTON_EXPORT ) );

        VerticalLayout subLayout = new VerticalLayout( exportStatusLabel );
        if ( dTable.getTableExporterConfig()
                   .isRemovedOrdersFromColumnsEnabled() )
        {
            subLayout.addComponent( removeOrdersFromColumns );
        }
        if ( dTable.getTableExporterConfig()
                   .isSelectHiddenColumnsEnabled() )
        {
            subLayout.addComponent( includeHiddenColumns );
        }

        HorizontalLayout mainLayout = new HorizontalLayout( subLayout );
        mainLayout.addStyleName( "context" );
        mainLayout.setWidth( 100, Unit.PERCENTAGE );
        mainLayout.setMargin( true );
        mainLayout.setSpacing( true );

        HorizontalLayout buttonsLayout = new HorizontalLayout( exportButton, closeButton );
        buttonsLayout.setComponentAlignment( exportButton, Alignment.MIDDLE_RIGHT );
        buttonsLayout.setComponentAlignment( closeButton, Alignment.MIDDLE_LEFT );
        buttonsLayout.setMargin( true );
        buttonsLayout.setSpacing( true );

        VerticalLayout windowLayout = new VerticalLayout( mainLayout, buttonsLayout );
        windowLayout.setComponentAlignment( buttonsLayout, Alignment.BOTTOM_CENTER );
        windowLayout.setMargin( false );
        windowLayout.setSpacing( false );
        windowLayout.addStyleName( "v-dtable-dialog" );
        setContent( windowLayout );

        exportStatus = DISABLED;
    }

    private void exportStatusWatcher()
    {
        if ( exportStatus != DISABLED )
        {
            if ( getExportedStatus().compareTo( EXPORT_ERROR ) == 0 )
            {
                exportStatusLabel.setValue( dTable.getTranslation()
                                                  .get( dTable.getTableConfig()
                                                              .getLocale(),
                                                        TranslatedWords.IN_PREPARING_DATA_AN_ERROR_OCCURRED ) );
                exportButton.setVisible( false );
                closeButton.setEnabled( true );

                exportStatus = DISABLED;
                UI.getCurrent()
                  .setPollInterval( DISABLING );
            }
            else if ( getExportedStatus().compareTo( EXPORT_STOP ) == 0 )
            {
                setExportedStatus( EXPORT_STOPPED );
            }
            else if ( getExportedStatus().compareTo( EXPORT_STOPPED ) == 0 )
            {
                exportStatus = DISABLED;
                exportStatusLabel.setValue( dTable.getTranslation()
                                                  .get( dTable.getTableConfig()
                                                              .getLocale(),
                                                        TranslatedWords.PREPARING_DATA_IS_STOPPED ) );
                exportButton.setVisible( false );
                closeButton.setEnabled( true );
            }
            else if ( EXPORT_BEGIN <= getExportedStatus() && getExportedStatus() < EXPORT_COMPLETE )
            {
                if ( EXPORT_BEGIN < getExportedStatus() )
                {
                    exportStatusLabel.setValue( dTable.getTranslation()
                                                      .get( dTable.getTableConfig()
                                                                  .getLocale(),
                                                            TranslatedWords.PREPARING_DATA )
                        + " (" + exportStatusFormatter.format( exportedStatus ) + ")..." );
                }
            }
            else if ( getExportedStatus().compareTo( EXPORT_COMPLETE ) == 0 )
            {
                exportButton.setEnabled( false );
                exportStatusLabel.setValue( dTable.getTranslation()
                                                  .get( dTable.getTableConfig()
                                                              .getLocale(),
                                                        TranslatedWords.PREPARING_DATA_IS_COMPLETED ) );
                setExportedStatus( EXPORT_COMPLETED );
            }
            else if ( getExportedStatus().compareTo( EXPORT_COMPLETED ) == 0 )
            {
                exportStatusLabel.setValue( dTable.getTranslation()
                                                  .get( dTable.getTableConfig()
                                                              .getLocale(),
                                                        TranslatedWords.PREPARING_DATA_FOR_DOWNLOAD ) );
                setExportedStatus( EXPORT_PREPARE );
            }
            else if ( getExportedStatus().compareTo( EXPORT_PREPARE ) == 0 )
            {
                prepareExportedFile();
            }
            else if ( getExportedStatus().compareTo( EXPORT_PREPARED ) == 0 )
            {
                exportButton.setEnabled( true );
                closeButton.setEnabled( true );
                exportStatusLabel.setValue( dTable.getTranslation()
                                                  .get( dTable.getTableConfig()
                                                              .getLocale(),
                                                        TranslatedWords.DATA_ARE_READY_FOR_DOWNLOAD ) );
                exportButton.setCaption( dTable.getTranslation()
                                               .get( dTable.getTableConfig()
                                                           .getLocale(),
                                                     TranslatedWords.DOWNLOAD_THE_DATA ) );

                exportStatus = DISABLED;
                UI.getCurrent()
                  .setPollInterval( DISABLING );
            }
        }
        else if ( exportStatus == DISABLED && UI.getCurrent()
                                                .getPollInterval() == DISABLING )
        {
            UI.getCurrent()
              .setPollInterval( DISABLED );
        }
    }

    /**
     * Export process runnable class.
     */
    private class ExportProcessRunner
        extends Thread
    {
        @Override
        public void run()
        {
            try
            {
                exportDataIntoFile();
            }
            catch ( Exception e )
            {
                Logger.getLogger( DTable.class.getName() )
                      .log( Level.INFO, e.getLocalizedMessage(), e );
            }
        }
    }

    public Label getExportStatusLabel()
    {
        return exportStatusLabel;
    }

    /**
     * Open file for writing exported data into it.
     * 
     * @param name filename
     * @return true if file is successfully open, otherwise false
     */
    protected abstract boolean openFile( String name );

    /**
     * Write exported data to the file.
     * 
     * @param name the file name
     * @param completeStep step in percentage value which show process status of exporting data into the file
     * @param step step value for the complete step
     */
    protected abstract void writeToFile( String name, float completeStep, float step );

    /**
     * Close file.
     */
    protected abstract void closeFile();

    /**
     * Main process for exporting data into file.
     */
    protected abstract void exportDataIntoFile();

    /**
     * Delete temporary folder.
     * 
     * @param folder temporary folder name
     */
    private static void deleteTempFolder( File folder )
    {
        try
        {
            if ( folder.exists() )
            {
                File[] files = folder.listFiles();
                for ( int i = 0; i < files.length; i++ )
                {
                    if ( files[i].isDirectory() )
                    {
                        deleteTempFolder( files[i] );
                    }
                    else
                    {
                        files[i].delete();
                    }
                }
            }
            folder.delete();
        }
        catch ( Exception e )
        {
            Logger.getLogger( DTable.class.getName() )
                  .log( Level.INFO, e.getLocalizedMessage(), e );
        }
    }

    /**
     * Close window which show export status.
     */
    private void closeWindow()
    {
        exportStatus = DISABLED;
        setExportedStatus( EXPORT_NOT_STARTED );
        setVisible( false );

        if ( UI.getCurrent()
               .getPollInterval() != DISABLED )
        {
            UI.getCurrent()
              .setPollInterval( DISABLED );
        }
        UI.getCurrent()
          .removeWindow( this );
    }

    /**
     * Prepare file into which data will be imported.
     */
    private void prepareExportedFile()
    {
        // get filenames to be added to archive
        String nonCompressedFileName;
        List<File> archivedFiles = new ArrayList<>();
        switch ( getExportedType() )
        {
            case CSV:
                for ( String name : dTable.getTableExporterConfig()
                                          .getColumnsDefinitionName() )
                {
                    archivedFiles.add( getTempFile( name ) );
                }
                nonCompressedFileName = dTable.getTableExporterConfig()
                                              .getColumnsDefinitionName()[0];
                break;
            case MS_ACCESS:
            case MS_EXCEL:
            default:
                archivedFiles.add( getTempFile( dTable.getTableExporterConfig()
                                                      .getFilename() ) );
                nonCompressedFileName = dTable.getTableExporterConfig()
                                              .getBaseName();
                break;
        }
        // create archive
        try
        {
            float exportingStatus = 0f;

            CompressionType compressionType = dTable.getTableExporterConfig()
                                                    .getCompressionType();
            if ( CompressionType.NONE.equals( compressionType ) && 1 < archivedFiles.size() )
            {
                compressionType = CompressionType.ZIP;
            }
            archivedFile = new File( getExportedFile().getCanonicalPath() + compressionType.getExtension() );

            switch ( compressionType )
            {
                case TAR_BZ2:
                    exportingStatus = CompressingProcessor.tarBz2( archivedFile, archivedFiles );
                    break;
                case TAR_GZ:
                    exportingStatus = CompressingProcessor.tarGz( archivedFile, archivedFiles );
                    break;
                case ZIP:
                    exportingStatus = CompressingProcessor.zip( archivedFile, archivedFiles );
                    break;
                case NONE:
                default:
                    archivedFile = new File( getExportedFile().getCanonicalPath() );
                    getTempFile( nonCompressedFileName ).renameTo( archivedFile );
                    exportingStatus = EXPORT_PREPARED;
                    break;
            }
            setExportedStatus( exportingStatus );
        }
        catch ( Exception e )
        {
            archivedFile = null;
            Logger.getLogger( DTable.class.getName() )
                  .log( Level.INFO, e.getLocalizedMessage(), e );
            setExportedStatus( EXPORT_ERROR, e.getLocalizedMessage() );
        }
    }

    /**
     * Get file extension.
     * 
     * @return file extension
     */
    private String getExtension()
    {
        switch ( exportingType )
        {
            case MS_ACCESS:
                return getDTable().getTableExporterConfig()
                                  .getMsAccessVersion()
                                  .getExtension();
            case MS_EXCEL:
                return getDTable().getTableExporterConfig()
                                  .getMsExcelVersion()
                                  .getExtension();
            case CSV:
            default:
                return exportingType.getExtension();
        }
    }

    /**
     * Get exported file.
     * 
     * @return exported file
     */
    private File getExportedFile()
    {
        try
        {
            if ( !new File( tempFolder ).exists() )
            {
                new File( tempFolder ).mkdir();
            }

            String timestamp = "";
            if ( dTable.getTableExporterConfig()
                       .isTimestampEnabled() )
            {
                SimpleDateFormat timestampFormat = new SimpleDateFormat( dTable.getTableExporterConfig()
                                                                               .getTimestampFormat() );
                timestamp = "." + timestampFormat.format( new Date() );
            }
            return new File( tempFolder + File.separatorChar + dTable.getTableExporterConfig()
                                                                     .getFilename()
                + timestamp + getExtension() );
        }
        catch ( Exception e )
        {
            Logger.getLogger( DTable.class.getName() )
                  .log( Level.INFO, e.getLocalizedMessage(), e );
            return null;
        }
    }

    /**
     * Get column types from table data for exported data.
     * 
     * @param name unique name for exported data
     * @return column types for exported data
     */
    private Map<Integer, ExportingColumnType> getTableDataType()
    {
        Map<Integer, ExportingColumnType> dataTypes = new HashMap<>();
        for ( int column = 0; column < getDTable().getTableProperties()
                                                  .getAllColumns(); column++ )
        {
            switch ( getDTable().getTableConfig()
                                .getColumnType( column ) )
            {
                case BIG_DECIMAL:
                    dataTypes.put( column, ExportingColumnType.BIG_DECIMAL );
                    break;
                case BIG_INTEGER:
                    dataTypes.put( column, ExportingColumnType.BIG_INTEGER );
                    break;
                case BOOLEAN:
                    dataTypes.put( column, ExportingColumnType.BOOLEAN );
                    break;
                case DATE:
                    dataTypes.put( column, ExportingColumnType.DATE );
                    break;
                case DOUBLE:
                    dataTypes.put( column, ExportingColumnType.DOUBLE );
                    break;
                case DECIMAL_NUMBER:
                case FLOAT:
                    dataTypes.put( column, ExportingColumnType.FLOAT );
                    break;
                case INTEGER:
                case INTEGER_NUMBER:
                    dataTypes.put( column, ExportingColumnType.INTEGER );
                    break;
                case LONG:
                    dataTypes.put( column, ExportingColumnType.LONG );
                    break;
                case MEMO:
                    dataTypes.put( column, ExportingColumnType.MEMO );
                    break;
                case TIMESTAMP:
                    dataTypes.put( column, ExportingColumnType.TIMESTAMP );
                    break;
                case TEXT:
                    dataTypes.put( column, ExportingColumnType.TEXT );
                default:
                    break;
            }
        }
        return dataTypes;
    }

    /**
     * Get status if process for exporting data can continue.
     * 
     * @return status if process for exporting data can continue
     */
    protected boolean canContinue()
    {
        return EXPORT_STOP <= getExportedStatus() && !isExportInterrupted();
    }

    /**
     * Get calculated step for exported columns: if have too many columns then we have to export rows data incrementally
     * over the column groups.
     * 
     * @param name unique name for exported data
     * @return calculated step for exported columns
     */
    protected int getExportedStep( String name )
    {
        try
        {
            float step = Math.max( 1, 100000f / getAllColumns( name ) );
            if ( step < EXPORT_STEP )
            {
                return (int) step;
            }
            return Math.min( EXPORT_STEP * 10, EXPORT_STEP * ( (int) ( step / EXPORT_STEP ) ) );
        }
        catch ( Exception e )
        {
            Logger.getLogger( DTable.class.getName() )
                  .log( Level.INFO, e.getLocalizedMessage(), e );
            return 0;
        }
    }

    protected boolean prepareExportedData( String name )
    {
        if ( dTable.prepareExportedData( name ) )
        {
            Map<Integer, ExportingColumnType> columnsTableDataType = getTableDataType();
            Map<Integer, ColumnDefinition> definitions = dTable.getTableExporterConfig()
                                                               .getColumnDefinition( name );

            if ( dTable.getTableExporterConfig()
                       .getBaseName()
                       .equals( name )
                && ( !dTable.getTableExporterConfig()
                            .isPrepareDataWithHiddenColumnsEnabled()
                    || !dTable.getTableExporterConfig()
                              .isPrepareDataWithRemovedColumnsEnabled() ) )
            {
                int[] removedColumns = dTable.getTableProperties()
                                             .getHiddenAndRemovedColumns( !dTable.getTableExporterConfig()
                                                                                 .isPrepareDataWithHiddenColumnsEnabled(),
                                                                          !dTable.getTableExporterConfig()
                                                                                 .isPrepareDataWithRemovedColumnsEnabled() );

                if ( removedColumns != null && 0 < removedColumns.length )
                {
                    for ( int column : removedColumns )
                    {
                        if ( definitions.containsKey( column ) )
                        {
                            definitions.remove( column );
                        }
                    }
                }
            }

            for ( int column : definitions.keySet() )
            {
                String header = dTable.getTableExporterConfig()
                                      .getColumnDefinitionHeader( name, column );
                Formatting formatting = dTable.getTableExporterConfig()
                                              .getColumnDefinitionFormatting( name, column );
                ExportingColumnType columnExportedType = getDTable().getTableExporterConfig()
                                                                    .getColumnDefinitionExportingType( name, column );
                Integer scale = getDTable().getTableExporterConfig()
                                           .getColumnDefinitionScale( name, column );
                Integer length = dTable.getTableExporterConfig()
                                       .getColumnDefinitionLength( name, column );

                // check column name
                if ( header == null || header.isEmpty() )
                {
                    header = "HEADER " + column;
                }

                // check data type
                switch ( getExportedType() )
                {
                    case MS_ACCESS:
                        // set text max length
                        if ( ExportingColumnType.TEXT.equals( columnExportedType ) )
                        {
                            if ( length == null )
                            {
                                length = Integer.valueOf( JetFormat.TEXT_FIELD_MAX_LENGTH );
                            }
                            length = Math.min( length, JetFormat.TEXT_FIELD_MAX_LENGTH );
                        }

                        // check column type
                        if ( columnExportedType != null && !columnsTableDataType.get( column )
                                                                                .equals( columnExportedType ) )
                        {
                            setExportedStatus( EXPORT_ERROR, getDTable().getTranslation()
                                                                        .get( dTable.getTableConfig()
                                                                                    .getLocale(),
                                                                              TranslatedWords.EXPORTED_COLUMN_DATA_TYPE_AND_COLUMN_DEFINITION_TYPE_ARE_NOT_THE_SAME ) );
                            return false;
                        }
                        break;
                    default:
                        break;
                }

                ColumnDefinition definition =
                    new ColumnDefinition( header, formatting, columnExportedType, length, scale );
                definitions.put( column, definition );
            }

            // write back to original
            dTable.getTableExporterConfig()
                  .setColumnDefinition( name, definitions );
            return true;
        }
        else
        {
            return false;
        }
    }

    protected long getAllRows( String name )
    {
        return dTable.getExportedAllRows( name );
    }

    protected int getAllColumns( String name )
    {
        return dTable.getTableExporterConfig()
                     .getColumnDefinition( name )
                     .size();
    }

    protected File getTempFile( String name )
    {
        try
        {
            if ( !new File( tempFolder ).exists() )
            {
                new File( tempFolder ).mkdir();
            }
            return new File( tempFolder + File.separatorChar + name + getExtension() );
        }
        catch ( Exception e )
        {
            Logger.getLogger( DTable.class.getName() )
                  .log( Level.INFO, e.getLocalizedMessage(), e );
            return null;
        }
    }

    protected String getCharset()
    {
        return dTable.getTableExporterConfig()
                     .getCharset()
                     .getCharset();
    }

    protected Object[] getOriginalData( String name, int startRow, int rows )
    {
        int allColumns = getAllColumns( name );
        Object[] data = dTable.getExportedData( name, startRow, rows );
        return Arrays.copyOfRange( data, 0, rows * allColumns );
    }

    protected Object[] getFromattedData( String name, int startRow, int rows )
    {
        int allColumns = getAllColumns( name );
        Object[] data = getOriginalData( name, startRow, rows );
        if ( data != null && 0 < data.length && !ExportedType.MS_ACCESS.equals( getExportedType() )
            && dTable.getTableExporterConfig()
                     .isFormattedDataEnabled() )
        {
            for ( int row = 0; row < rows; row++ )
            {
                List<Integer> columns = new ArrayList<>( getDTable().getTableExporterConfig()
                                                                    .getColumnDefinition( name )
                                                                    .keySet() );
                Collections.sort( columns );
                for ( int column = 0; column < allColumns; column++ )
                {
                    Formatting formatting = dTable.getTableExporterConfig()
                                                  .getColumnDefinitionFormatting( name, columns.get( column ) );
                    data[row * allColumns + column] =
                        FormattingProcessor.process( formatting, data[row * allColumns + column] );
                }
            }
        }
        return Arrays.copyOfRange( data, 0, rows * allColumns );
    }

    protected DTable getDTable()
    {
        return dTable;
    }

    protected boolean isExportInterrupted()
    {
        return exportInterrupted;
    }

    protected Float getExportedStatus()
    {
        return exportedStatus;
    }

    protected void setExportedStatus( float exportedStatus, String description )
    {
        setExportedStatus( exportedStatus );
        if ( description != null && !description.isEmpty() )
        {
            new ErrorUtil( dTable.getTranslation()
                                 .get( dTable.getTableConfig()
                                             .getLocale(),
                                       TranslatedWords.ERROR ),
                           description ).show();
        }
    }

    protected void setExportedStatus( Float exportedStatus )
    {
        if ( canContinue() )
        {
            this.exportedStatus = exportedStatus;
        }
        else if ( EXPORT_STOP <= getExportedStatus() )
        {
            this.exportedStatus = EXPORT_STOPPED;
        }
        else
        {
            this.exportedStatus = exportedStatus;
        }

        if ( this.exportedStatus.compareTo( EXPORT_ERROR ) == 0 || this.exportedStatus.compareTo( EXPORT_STOPPED ) == 0
            || this.exportedStatus.compareTo( EXPORT_COMPLETED ) == 0 )
        {
            dTable.exportProcessCompleted();
        }
    }

    protected ExportedType getExportedType()
    {
        return exportingType;
    }

    protected void setExportedType( ExportedType exportingType )
    {
        this.exportingType = exportingType;
    }

}
